@@ -10,22 +10,25 @@ Usage:
 Commands:
 {{ command_list }}
 Options:
-  --host=<host>        OneP URL. Default is $EXO_HOST or m2.exosite.com
-  --port=<port>        OneP port. Default is $EXO_HOST or 80
+  --host=<host>        OneP host. Default is $EXO_HOST or m2.exosite.com
+  --port=<port>        OneP port. Default is $EXO_PORT or 443
   --httptimeout=<sec>  HTTP timeout [default: 60]
   --https              Enable HTTPS (deprecated, HTTPS is default)
   --http               Disable HTTPS
   --debug              Show info like stack traces
   --debughttp          Turn on debug level logging in pyonep
   --discreet           Obfuscate RIDs in stdout and stderr
+  -c --clearcache     Invalidate Portals cache after running command
+  --portals=<server>   Portals server [default: https://portals.exosite.com]
   -h --help            Show this screen
   -v --version         Show version
 
 See 'exo <command> --help' for more information on a specific command.
 """
 
-# Copyright (c) 2013, Exosite, LLC
+# Copyright (c) 2014, Exosite, LLC
 # All rights reserved
+from __future__ import unicode_literals
 import sys
 import os
 import json
@@ -38,7 +41,11 @@ from pprint import pprint
 from operator import itemgetter
 import logging
 from collections import defaultdict
-import StringIO
+
+import six
+from six import StringIO
+from six import iteritems
+
 import pytz
 # python 2.6 support
 try:
@@ -159,24 +166,7 @@ Options:
         '''Update a resource from a json description passed on stdin.\n\nUsage:
     exo [options] update <cik> (<rid> - | -)
 
-
-Restrictions:
-
-Client Description
-    Resource limits must not be lowered below current use level. Resources must be
-    dropped prior to lowering the limits. For daily limits, those may be lowered at
-    any point and take immediate affect.
-
-Dataport Description
-    Format must not be changed.
-
-Datarule Descriptions
-    Format must not be changed.
-
-Dispatch Description
-    If the recipient or method is changed, and the recipient/method combination has
-    never been used before, then further dispatches will be halted until a
-    Validation Request is sent and validated.'''),
+    For details see https://github.com/exosite/api/tree/master/rpc#update'''),
     ('map',
         '''Add an alias to a resource.\n\nUsage:
     exo [options] map <cik> <rid> <alias>'''),
@@ -257,7 +247,16 @@ Options:
         write/read API is used. Writes are executed before reads.'''),
     ('activate', '''Activate a model-backed device, based on vendor name,
     vendor model and serial number.\n\nUsage:
-     exo [options] activate <vendor> <model> <sn>''')
+    exo [options] activate <vendor> <model> <sn>'''),
+    ('portals', '''Invalidate the Portals cache for a CIK by telling Portals
+a particular procedure was taken on client identified by <cik>.\n\nUsage:
+    exo [options] portals clearcache <cik> [<procedure> ...]
+
+    <procedure> may be any of:
+    activate, create, deactivate, drop, map, revoke, share, unmap, update
+
+    If no <procedure> is specified, Exoline tells Portals that all of the
+    procedures on the list were performed on the client.''')
     ])
 
 # shared sections of documentation
@@ -272,7 +271,7 @@ doc_replace = {
     If timezone information is omitted, local timezone is assumed
     If time part is omitted, it assumes 00:00:00.
     To report through the present time, omit --end or pass --end=now''',
-    '{{ helpoption }}': '''-h --help  Show this screen.''',
+    '{{ helpoption }}': '''    -h --help  Show this screen.''',
 }
 
 # load plugins. use timezone because this file may be running
@@ -308,7 +307,8 @@ for k in cmd_doc:
 
 
 class ExolineOnepV1(onep.OnepV1):
-    '''Subclass that re-adds deprecated commands used by Portals'''
+    '''Subclass that re-adds deprecated commands needed for devices created
+    in Portals before the commands were deprecated.'''
     def comment(self, cik, rid, visibility, comment, defer=False):
         return self._call('comment', cik, [rid, visibility, comment], defer)
 
@@ -326,7 +326,8 @@ class ExoRPC():
                  port=None,
                  httptimeout=60,
                  https=False,
-                 verbose=True):
+                 verbose=True,
+                 logrequests=False):
 
         if port is None:
             port = DEFAULT_PORT_HTTPS if https else DEFAULT_PORT
@@ -335,7 +336,8 @@ class ExoRPC():
                                httptimeout=httptimeout,
                                https=https,
                                agent="Exoline {0}".format(__version__),
-                               reuseconnection=True)
+                               reuseconnection=True,
+                               logrequests=logrequests)
 
     def _raise_for_response(self, isok, response, call=None):
         if not isok:
@@ -772,15 +774,15 @@ class ExoRPC():
         else:
             desc = typ + ' ' + id
 
-        print(u'{0}{1}{2} {3} {4}'.format(
+        print('{0}{1}{2} {3} {4}'.format(
             spacer,
             name,
             ' ' * (max_name - len(name)),
             desc,
-            u'' if len(opt) == 0 else u'({0})'.format(u', '.join(
-                [u'{0}: {1}'.format(k, v) for k, v in opt.iteritems()]))))
+            '' if len(opt) == 0 else '({0})'.format(', '.join(
+                ['{0}: {1}'.format(k, v) for k, v in iteritems(opt)]))))
 
-    def tree(self, cik, aliases=None, cli_args={}, spacer=u'', level=0, info_options={}):
+    def tree(self, cik, aliases=None, cli_args={}, spacer='', level=0, info_options={}):
         '''Print a tree of entities in OneP'''
         max_level = int(cli_args['--level'])
         # print root node
@@ -817,26 +819,26 @@ class ExoRPC():
             # _listing_with_info(): [{'<rid0>':<info0>, '<rid1>':<info1>},
             #                       {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]
         except pyonep.exceptions.OnePlatformException:
-            print(spacer + u"  └─listing for {0} failed. Is info['basic']['status'] == 'expired'?".format(cik))
+            print(spacer + "  └─listing for {0} failed. info['basic']['status'] is probably not valid.".format(cik))
         else:
             # calculate the maximum length name of all children
-            lengths = [len(l[1]['description']['name']) for i in range(len(types)) for l in listing[i].iteritems()]
+            lengths = [len(l[1]['description']['name']) for i in range(len(types)) for l in iteritems(listing[i])]
             max_name = 0 if len(lengths) == 0 else max(lengths)
 
             # print everything
             for t_idx, t in enumerate(types):
-                typelisting = OrderedDict(sorted(listing[t_idx].iteritems(), key=lambda x: x[1]['description']['name'].lower()))
+                typelisting = OrderedDict(sorted(iteritems(listing[t_idx]), key=lambda x: x[1]['description']['name'].lower()))
                 islast_nonempty_type = (t_idx == len(types) - 1) or (all(len(x) == 0 for x in listing[t_idx + 1:]))
                 for rid_idx, rid in enumerate(typelisting):
                     info = typelisting[rid]
                     islastoftype = rid_idx == len(typelisting) - 1
                     islast = islast_nonempty_type and islastoftype
                     if islast:
-                        child_spacer = spacer + u'    '
-                        own_spacer   = spacer + u'  └─'
+                        child_spacer = spacer + '    '
+                        own_spacer   = spacer + '  └─'
                     else:
-                        child_spacer = spacer + u'  │ '
-                        own_spacer   = spacer + u'  ├─'
+                        child_spacer = spacer + '  │ '
+                        own_spacer   = spacer + '  ├─'
 
                     if t == 'client':
                         next_cik = info['key']
@@ -1305,6 +1307,39 @@ class ExoProvision():
         self.raise_for_status(r)
         return r.text
 
+class ExoPortals():
+    '''Provides access to the Portals APIs'''
+
+    # list of procedures that may be included in invalidation data
+    writeprocs = ['activate',
+                  'create',
+                  'deactivate',
+                  'drop',
+                  'map',
+                  'revoke',
+                  'share',
+                  'unmap',
+                  'update']
+
+    def __init__(self, portalsserver='https://portals.exosite.com'):
+        self.portalsserver = portalsserver
+
+    def invalidate(self, data):
+        # This API is documented here:
+        # https://i.exosite.com/display/DEVPORTALS/Portals+Cache+Invalidation+API
+        data = json.dumps(data)
+        #print('invalidating with ' + data)
+        try:
+            response = requests.post(self.portalsserver + '/api/portals/v1/cache',
+                                     data=data)
+        except Exception as ex:
+            raise ExoException('Failed to connect to ' + self.portalsserver)
+        try:
+            response.raise_for_status
+        except Exception as ex:
+            raise ExoException('Bad status from Portals cache invalidate API call: ' + ex)
+
+
 def parse_ts(s):
     return parse_ts_tuple(parser.parse(s).timetuple())
 
@@ -1339,9 +1374,9 @@ def format_time(sec):
                     [60, 'm']]
     text = ""
     for s, label in intervals:
-        if sec >= s and sec / s > 0:
-            text = "{0} {1}{2}".format(text, sec / s, label)
-            sec -= s * (sec / s)
+        if sec >= s and sec // s > 0:
+            text = "{0} {1}{2}".format(text, sec // s, label)
+            sec -= s * (sec // s)
     if sec > 0:
         text += " {0}s".format(sec)
     return text.strip()
@@ -1369,7 +1404,7 @@ def spark(numbers, empty_val=None):
 
     for number in numbers:
         if number == empty_val:
-            out.append(u" ")
+            out.append(" ")
         else:
             if (number - min_value) != 0 and value_scale != 0:
                 scaled_value = (number - min_value) / value_scale
@@ -1387,7 +1422,11 @@ def spark(numbers, empty_val=None):
             elif num == 7:
                 num = 6
 
-            out.append(unichr(int(9601 + num)))
+            if six.PY3:
+                unichrfn = chr
+            else:
+                unichrfn = unichr
+            out.append(unichrfn(int(9601 + num)))
 
     return ''.join(out)
 
@@ -1410,7 +1449,7 @@ def show_intervals(er, cik, rid, start, end, limit, numstd=None):
 
     if len(data) == 0:
         return
-    intervals = [data[i - 1][0] - data[i][0] for i in xrange(1, len(data))]
+    intervals = [data[i - 1][0] - data[i][0] for i in range(1, len(data))]
     intervals = sorted(intervals)
 
     if numstd is not None:
@@ -1434,9 +1473,12 @@ def show_intervals(er, cik, rid, start, end, limit, numstd=None):
         else:
             critfn = lambda x: bin_min <= x and x <= bin_max
         #bins.append((bin_min, bin_max, float(
-        #    sum(itertools.imap(critfn, intervals)))))
-        bins.append(float(
-            sum(itertools.imap(critfn, intervals))))
+        #    sum(map(critfn, intervals)))))
+        if six.PY3:
+            mapfn = map
+        else:
+            mapfn = itertools.imap
+        bins.append(float(sum(mapfn(critfn, intervals))))
 
     if False:
         # debug
@@ -1504,19 +1546,19 @@ def read_cmd(er, cik, rids, args):
         # default to UTC
         try:
             tz = timezone.localtz()
-        except pytz.UnknownTimeZoneError, e:
-            print "Unable to detect local time zone, defaulting to UTC"
+        except pytz.UnknownTimeZoneError as e:
+            print('Unable to detect local time zone, defaulting to UTC')
             tz = pytz.utc
     else:
         try:
             tz = pytz.timezone(tz)
-        except Exception, e:
+        except Exception as e:
             #default to utc if error
-            print "Error parsing --tz option, defaulting to local timezone"
+            print('Error parsing --tz option, defaulting to local timezone')
             try:
                 tz = timezone.localtz()
-            except pytz.UnknownTimeZoneError, e:
-                print "Unable to detect local time zone, defaulting to UTC"
+            except pytz.UnknownTimeZoneError as e:
+                print('Unable to detect local time zone, defaulting to UTC')
                 tz = pytz.utc
 
     recarriage = re.compile('\r(?!\\n)')
@@ -1609,10 +1651,13 @@ def pretty_print(arg):
     print(json.dumps(arg, sort_keys=True, indent=4, separators=(',', ': ')))
 
 
-
-
 def handle_args(cmd, args):
-    er = ExoRPC(host=args['--host'], port=args['--port'], https=not args['--http'], httptimeout=args["--httptimeout"])
+    use_https = False if args['--http'] is True else True
+    er = ExoRPC(host=args['--host'],
+                port=args['--port'],
+                https=use_https,
+                httptimeout=args['--httptimeout'],
+                logrequests=args['--clearcache'])
     if cmd in ['ip', 'data']:
         if args['--https'] is True or args['--port'] is not None or args['--debughttp'] is True:
             # TODO: support these
@@ -1622,8 +1667,12 @@ def handle_args(cmd, args):
     if cmd in ['activate']:
         if args['--https'] is True or args['--port'] is not None or args['--debughttp'] is True:
             # TODO: support these
-            raise ExoException('--https, --port, and --debughttp are not supported for ip and data commands.')
+            raise ExoException('--https, --port, and --debughttp are not supported for provisioning commands.')
         ep = ExoProvision(url='http://' + args['--host'])
+
+    if cmd in ['portals'] or args['--clearcache']:
+        portals = ExoPortals(args['--portals'])
+
     if '<cik>' in args and args['<cik>'] is not None:
         cik = args['<cik>']
         if type(cik) is list:
@@ -1657,244 +1706,267 @@ def handle_args(cmd, args):
     else:
         pr = plain_print
 
-    if cmd == 'read':
-        read_cmd(er, cik, rids, args)
-    elif cmd == 'write':
-        er.write(cik, rids[0], args['--value'])
-    elif cmd == 'record':
-        interval = args['--interval']
-        if interval is None:
-            entries = []
-            # split timestamp, value
-            has_errors = False
-            if args['-']:
-                headers = ['timestamp', 'value']
-                dr = csv.DictReader(sys.stdin, headers)
-                for row in dr:
-                    # TODO: handle something other than unix timestamps
-                    entries.append([int(row['timestamp']), row['value']])
-            else:
-                tvalues = args['--value']
-                reentry = re.compile('(-?\d+),(.*)')
-                for tv in tvalues:
-                    match = reentry.match(tv)
-                    if match is None:
-                        try:
-                            t, v = tv.split(',')
-                            t = parse_ts(t)
-                            entries.append([t, v])
-                        except Exception:
-                            sys.stderr.write(
-                                'Line not in <timestamp>,<value> format: {0}'.format(tv))
-                            has_errors = True
-                    else:
-                        g = match.groups()
+    try:
+        if cmd == 'read':
+            read_cmd(er, cik, rids, args)
+        elif cmd == 'write':
+            er.write(cik, rids[0], args['--value'])
+        elif cmd == 'record':
+            interval = args['--interval']
+            if interval is None:
+                entries = []
+                # split timestamp, value
+                has_errors = False
+                if args['-']:
+                    headers = ['timestamp', 'value']
+                    dr = csv.DictReader(sys.stdin, headers)
+                    for row in dr:
                         # TODO: handle something other than unix timestamps
-                        entries.append([int(g[0]), g[1]])
+                        entries.append([int(row['timestamp']), row['value']])
+                else:
+                    tvalues = args['--value']
+                    reentry = re.compile('(-?\d+),(.*)')
+                    for tv in tvalues:
+                        match = reentry.match(tv)
+                        if match is None:
+                            try:
+                                t, v = tv.split(',')
+                                t = parse_ts(t)
+                                entries.append([t, v])
+                            except Exception:
+                                sys.stderr.write(
+                                    'Line not in <timestamp>,<value> format: {0}'.format(tv))
+                                has_errors = True
+                        else:
+                            g = match.groups()
+                            # TODO: handle something other than unix timestamps
+                            entries.append([int(g[0]), g[1]])
 
-            if has_errors or len(entries) == 0:
-                raise ExoException("Problems with input.")
+                if has_errors or len(entries) == 0:
+                    raise ExoException("Problems with input.")
+                else:
+                    er.record(cik, rids[0], entries)
             else:
-                er.record(cik, rids[0], entries)
-        else:
+                if args['-']:
+                    values = [v.strip() for v in sys.stdin.readlines()]
+                else:
+                    values = args['--value']
+                interval = int(interval)
+                if interval <= 0:
+                    raise ExoException("--interval must be positive")
+                er.record_backdate(cik, rids[0], interval, values)
+        elif cmd == 'create':
+            typ = args['--type']
+            ridonly = args['--ridonly']
+            cikonly = args['--cikonly']
+            if ridonly and cikonly:
+                raise ExoException('--ridonly and --cikonly are mutually exclusive')
             if args['-']:
-                values = [v.strip() for v in sys.stdin.readlines()]
+                s = sys.stdin.read()
+                try:
+                    desc = json.loads(s)
+                except Exception as ex:
+                    raise ExoException(ex)
+                rid = er.create(cik,
+                                type=typ,
+                                desc=desc,
+                                name=args['--name'])
+            elif typ == 'client':
+                rid = er.create_client(cik,
+                                    name=args['--name'])
+            elif typ == 'dataport':
+                rid = er.create_dataport(cik,
+                                        args['--format'],
+                                        name=args['--name'])
             else:
-                values = args['--value']
-            interval = int(interval)
-            if interval <= 0:
-                raise ExoException("--interval must be positive")
-            er.record_backdate(cik, rids[0], interval, values)
-    elif cmd == 'create':
-        typ = args['--type']
-        ridonly = args['--ridonly']
-        cikonly = args['--cikonly']
-        if ridonly and cikonly:
-            raise ExoException('--ridonly and --cikonly are mutually exclusive')
-        if args['-']:
+                raise ExoException('No defaults for {0}.'.format(args['--type']))
+            if ridonly:
+                pr(rid)
+            elif cikonly:
+                print(er.info(cik, rid, cikonly=True))
+            else:
+                pr('rid: {0}'.format(rid))
+                if typ == 'client':
+                    # for convenience, look up the cik
+                    print('cik: {0}'.format(er.info(cik, rid, cikonly=True)))
+            if args['--alias'] is not None:
+                er.map(cik, rid, args['--alias'])
+                if not ridonly:
+                    print("alias: {0}".format(args['--alias']))
+        elif cmd == 'update':
             s = sys.stdin.read()
             try:
                 desc = json.loads(s)
             except Exception as ex:
                 raise ExoException(ex)
-            rid = er.create(cik,
-                            type=typ,
-                            desc=desc,
-                            name=args['--name'])
-        elif typ == 'client':
-            rid = er.create_client(cik,
-                                   name=args['--name'])
-        elif typ == 'dataport':
-            rid = er.create_dataport(cik,
-                                     args['--format'],
-                                     name=args['--name'])
-        else:
-            raise ExoException('No defaults for {0}.'.format(args['--type']))
-        if ridonly:
-            pr(rid)
-        elif cikonly:
-            print(er.info(cik, rid, cikonly=True))
-        else:
-            pr('rid: {0}'.format(rid))
-            if typ == 'client':
-                # for convenience, look up the cik
-                print('cik: {0}'.format(er.info(cik, rid, cikonly=True)))
-        if args['--alias'] is not None:
-            er.map(cik, rid, args['--alias'])
-            if not ridonly:
-                print("alias: {0}".format(args['--alias']))
-    elif cmd == 'update':
-        s = sys.stdin.read()
-        try:
-            desc = json.loads(s)
-        except Exception as ex:
-            raise ExoException(ex)
-        pr(er.update(cik, rids[0], desc=desc))
-    elif cmd == 'map':
-        er.map(cik, rids[0], args['<alias>'])
-    elif cmd == 'unmap':
-        er.unmap(cik, args['<alias>'])
-    elif cmd == 'lookup':
-        # look up by cik or alias
-        cik_to_find = args['--cik']
-        if cik_to_find is not None:
-            cik_to_find = er.lookup_shortcut(cik_to_find)
-            rid = er.lookup_rid(cik, cik_to_find)
-            if rid is not None:
-                pr(rid)
-        else:
-            alias = args['<alias>']
-            if alias is None:
-                alias = ""
-            pr(er.lookup(cik, alias))
-    elif cmd == 'drop':
-        if args['--all-children']:
-            er.drop_all_children(cik)
-        else:
-            if len(rids) == 0:
-                raise ExoException("<rid> is required")
-            er.drop(cik, rids)
-    elif cmd == 'listing':
-        types = args['--type']
-        if len(types) == 0:
-            types = ['client', 'dataport', 'datarule', 'dispatch']
-        listing = er.listing(cik, types)
-        if args['--plain']:
-            for l in listing:
-                for rid in listing[0]:
-                    print(rid)
-        else:
-            pr(json.dumps(listing))
-    elif cmd == 'info':
-        include = args['--include']
-        include = [] if include is None else [key.strip()
-                                              for key in include.split(',')]
-        exclude = args['--exclude']
-        exclude = [] if exclude is None else [key.strip()
-                                              for key in exclude.split(',')]
+            pr(er.update(cik, rids[0], desc=desc))
+        elif cmd == 'map':
+            er.map(cik, rids[0], args['<alias>'])
+        elif cmd == 'unmap':
+            er.unmap(cik, args['<alias>'])
+        elif cmd == 'lookup':
+            # look up by cik or alias
+            cik_to_find = args['--cik']
+            if cik_to_find is not None:
+                cik_to_find = er.lookup_shortcut(cik_to_find)
+                rid = er.lookup_rid(cik, cik_to_find)
+                if rid is not None:
+                    pr(rid)
+            else:
+                alias = args['<alias>']
+                if alias is None:
+                    alias = ""
+                pr(er.lookup(cik, alias))
+        elif cmd == 'drop':
+            if args['--all-children']:
+                er.drop_all_children(cik)
+            else:
+                if len(rids) == 0:
+                    raise ExoException("<rid> is required")
+                er.drop(cik, rids)
+        elif cmd == 'listing':
+            types = args['--type']
+            if len(types) == 0:
+                types = ['client', 'dataport', 'datarule', 'dispatch']
+            listing = er.listing(cik, types)
+            if args['--plain']:
+                for l in listing:
+                    for rid in listing[0]:
+                        print(rid)
+            else:
+                pr(json.dumps(listing))
+        elif cmd == 'info':
+            include = args['--include']
+            include = [] if include is None else [key.strip()
+                                                for key in include.split(',')]
+            exclude = args['--exclude']
+            exclude = [] if exclude is None else [key.strip()
+                                                for key in exclude.split(',')]
+
+            options = er.make_info_options(include, exclude)
+            info = er.info(cik,
+                        rids[0],
+                        options=options,
+                        cikonly=args['--cikonly'],
+                        recursive=args['--recursive'])
+            if args['--pretty']:
+                pr(info)
+            else:
+                # output json
+                pr(json.dumps(info))
+        elif cmd == 'flush':
+            er.flush(cik, rids)
+        elif cmd == 'usage':
+            allmetrics = ['client',
+                        'dataport',
+                        'datarule',
+                        'dispatch',
+                        'email',
+                        'http',
+                        'sms',
+                        'xmpp']
+
+            start, end = get_startend(args)
+            er.usage(cik, rids[0], allmetrics, start, end)
+        # special commands
+        elif cmd == 'tree':
+            er.tree(cik, cli_args=args)
+        elif cmd == 'script':
+            # cik is a list of ciks
+            er.upload_script(cik, args['<script-file>'],
+                            name=args['--name'],
+                            recursive=args['--recursive'],
+                            create=args['--create'])
+        elif cmd == 'spark':
+            days = int(args['--days'])
+            end = parse_ts_tuple(datetime.now().timetuple())
+            start = parse_ts_tuple((datetime.now() - timedelta(days=days)).timetuple())
+            numstd = args['--stddev']
+            numstd = int(numstd) if numstd is not None else None
+            show_intervals(er, cik, rids[0], start, end, limit=1000000, numstd=numstd)
+        elif cmd == 'copy':
+            destcik = args['<destination-cik>']
+            newrid, newcik = er.copy(cik, destcik)
+            if args['--cikonly']:
+                pr(newcik)
+            else:
+                pr('cik: ' + newcik)
+        elif cmd == 'diff':
+            if sys.version_info < (2, 7):
+                raise ExoException('diff command requires Python 2.7 or above')
+
+            diffs = er.diff(cik,
+                            args['<cik2>'],
+                            full=args['--full'],
+                            nochildren=args['--no-children'])
+            if diffs is not None:
+                print(diffs)
+        elif cmd == 'ip':
+            pr(ed.ip())
+        elif cmd == 'data':
+            reads = args['--read']
+            writes = args['--write']
+            def get_alias_values(writes):
+                # TODO: support values with commas
+                alias_values = []
+                re_assign = re.compile('(.*),(.*)')
+                for w in writes:
+                    if w.count(',') > 1:
+                        raise ExoException('Values with commas not supported yet.')
+                    m = re_assign.match(w)
+                    if m is None or len(m.groups()) != 2:
+                        raise ExoException("Bad alias assignment format")
+                    alias_values.append(m.groups())
+                return alias_values
+
+            if len(reads) > 0 and len(writes) > 0:
+                alias_values = get_alias_values(writes)
+                print(ed.writeread(cik, alias_values, reads))
+            elif len(reads) > 0:
+                print(ed.read(cik, reads))
+            elif len(writes) > 0:
+                alias_values = get_alias_values(writes)
+                ed.write(cik, alias_values)
+        elif cmd == 'activate':
+            pr(ep.activate(args['<vendor>'], args['<model>'], args['<sn>']))
+        elif cmd == 'portals':
+
+            procedures = args['<procedure>']
+            if len(procedures) == 0:
+                procedures = ExoPortals.writeprocs
+            else:
+                unknownprocs = []
+                for p in procedures:
+                    if p not in ExoPortals.writeprocs:
+                        unknownprocs.append(p)
+                if len(unknownprocs) > 0:
+                    raise ExoException('Unknown procedure(s) {0}'.format(','.join(unknownprocs)))
+            data = {'auth': {'cik': cik}, 'calls':[{'procedure': p, 'arguments': [], 'id': i} for i, p in enumerate(procedures)]}
+            portals.invalidate(data)
 
-        options = er.make_info_options(include, exclude)
-        info = er.info(cik,
-                       rids[0],
-                       options=options,
-                       cikonly=args['--cikonly'],
-                       recursive=args['--recursive'])
-        if args['--pretty']:
-            pr(info)
         else:
-            # output json
-            pr(json.dumps(info))
-    elif cmd == 'flush':
-        er.flush(cik, rids)
-    elif cmd == 'usage':
-        allmetrics = ['client',
-                      'dataport',
-                      'datarule',
-                      'dispatch',
-                      'email',
-                      'http',
-                      'sms',
-                      'xmpp']
-
-        start, end = get_startend(args)
-        er.usage(cik, rids[0], allmetrics, start, end)
-    # special commands
-    elif cmd == 'tree':
-        er.tree(cik, cli_args=args)
-    elif cmd == 'script':
-        # cik is a list of ciks
-        er.upload_script(cik, args['<script-file>'],
-                         name=args['--name'],
-                         recursive=args['--recursive'],
-                         create=args['--create'])
-    elif cmd == 'spark':
-        days = int(args['--days'])
-        end = parse_ts_tuple(datetime.now().timetuple())
-        start = parse_ts_tuple((datetime.now() - timedelta(days=days)).timetuple())
-        numstd = args['--stddev']
-        numstd = int(numstd) if numstd is not None else None
-        show_intervals(er, cik, rids[0], start, end, limit=1000000, numstd=numstd)
-    elif cmd == 'copy':
-        destcik = args['<destination-cik>']
-        newrid, newcik = er.copy(cik, destcik)
-        if args['--cikonly']:
-            pr(newcik)
-        else:
-            pr('cik: ' + newcik)
-    elif cmd == 'diff':
-        if sys.version_info < (2, 7):
-            raise ExoException('diff command requires Python 2.7 or above')
-
-        diffs = er.diff(cik,
-                        args['<cik2>'],
-                        full=args['--full'],
-                        nochildren=args['--no-children'])
-        if diffs is not None:
-            print(diffs)
-    elif cmd == 'ip':
-        pr(ed.ip())
-    elif cmd == 'data':
-        reads = args['--read']
-        writes = args['--write']
-        def get_alias_values(writes):
-            # TODO: support values with commas
-            alias_values = []
-            re_assign = re.compile('(.*),(.*)')
-            for w in writes:
-                if w.count(',') > 1:
-                    raise ExoException('Values with commas not supported yet.')
-                m = re_assign.match(w)
-                if m is None or len(m.groups()) != 2:
-                    raise ExoException("Bad alias assignment format")
-                alias_values.append(m.groups())
-            return alias_values
-
-        if len(reads) > 0 and len(writes) > 0:
-            alias_values = get_alias_values(writes)
-            print(ed.writeread(cik, alias_values, reads))
-        elif len(reads) > 0:
-            print(ed.read(cik, reads))
-        elif len(writes) > 0:
-            alias_values = get_alias_values(writes)
-            ed.write(cik, alias_values)
-    elif cmd == 'activate':
-        pr(ep.activate(args['<vendor>'], args['<model>'], args['<sn>']))
-    else:
-        # search plugins
-        handled = False
-        for plugin in plugins:
-            if cmd in plugin.command():
-                options = {'cik': cik, 'rpc': er, 'exception': ExoException}
-                try:
-                    options['data'] = ed
-                except NameError:
-                    # no problem
-                    pass
-                plugin.run(cmd, args, options)
-                handled = True
-                break
-        if not handled:
-            raise ExoException("Command not handled")
+            # search plugins
+            handled = False
+            for plugin in plugins:
+                if cmd in plugin.command():
+                    options = {'cik': cik, 'rpc': er, 'exception': ExoException}
+                    try:
+                        options['data'] = ed
+                    except NameError:
+                        # no problem
+                        pass
+                    plugin.run(cmd, args, options)
+                    handled = True
+                    break
+            if not handled:
+                raise ExoException("Command not handled")
+    finally:
+        if args['--clearcache']:
+            for req in er.exo.loggedrequests():
+                procs = [c['procedure'] for c in req['calls']]
+                # if operation will invalidate the Portals cache...
+                if len([p for p in procs if p in ExoPortals.writeprocs]) > 0:
+                    portals.invalidate(req)
 
 
 class DiscreetFilter(object):
@@ -2006,13 +2078,13 @@ def run(argv, stdin=None):
     try:
         if stdin is None:
             stdin = sys.stdin
-        elif type(stdin) is str:
-            sio = StringIO.StringIO()
+        elif type(stdin) is str or type(stdin) is unicode:
+            sio = StringIO()
             sio.write(stdin)
             sio.seek(0)
             stdin = sio
-        stdout = StringIO.StringIO()
-        stderr = StringIO.StringIO()
+        stdout = StringIO()
+        stderr = StringIO()
 
         # unicode causes problems in docopt
         argv = [str(a) for a in argv]
